from copy import deepcopy
from dataclasses import dataclass, asdict
from .validation import validate_var
from .from_dict_generator import gen_from_any


@dataclass
class BaseModel:
  def __post_init__(self):
    self.__pre_validation__()
    self.validate_attributes()
    self.__post_validation__()
  
  def __pre_validation__(self): pass

  def __post_validation__(self): pass

  def to_dict(self):
    return asdict(self)

  def validate_attributes(self, raise_error=True, return_error_message=False):
    dc_fields = self.__dataclass_fields__
    for dc_field_key in dc_fields:
      if dc_fields[dc_field_key].init:
        ann = dc_fields[dc_field_key].type
        att = self.__getattribute__(dc_field_key)
        try:
          validate_var(att, ann)
        except Exception as e:
          error_message = f'({self.__class__.__name__}) -> attribute ({dc_field_key}) -> ' + e.args[0]
          if raise_error: raise TypeError(error_message)
          if return_error_message: return error_message
          return False
    
    if not raise_error: return True

  @classmethod
  def from_dict(cls, d: dict, raise_if_extra_attribute=False):
    if not isinstance(d, dict): raise TypeError('Input is not a dict!')
    d_copy = deepcopy(d)
    d_return = dict()

    dc_fields = cls.__dataclass_fields__  # type: ignore
    for key in d:
      if key not in dc_fields:
        if raise_if_extra_attribute:
          raise TypeError(f'{key} is not an attribute of {cls.__name__}')
      else:
        if dc_fields[key].init:
          try:
            d_return[key] = gen_from_any(d_copy[key], dc_fields[key].type)
          except Exception as e:
            error_message = f'({cls.__name__}) attribute ({key}) of type ' + e.args[0]
            raise TypeError(error_message)
      
    # try to create missing attributes (maybe they have default values)
    # for key in dc_fields:
    #   if key not in d_return:
    #     try:
    #       key_class = dc_fields[key].type
    #       d_return[key] = key_class()
    #     except:
    #       raise TypeError(f'{cls.__name__} attribute {key} is missing')

    return cls(**d_return) # type: ignore

    #list -> get_list=None, get_args=()
    #list[Any] -> get_origin=list, get_args=(Any,)
    #list[T] -> get_origin=list, get_args=(T,)
    #list[X, Y] -> Considerar só X



from typing import Any, Literal, Union, get_args, get_origin

def validate_var(var: Any, ann: Any) -> None:
  '''
  Validate if the variable type is the expected by the annotation
  '''
  origin = get_origin(ann)
  if origin is not None:
    return validate_with_origin(origin, var, get_args(ann))

  return validate_without_origin(var, ann)

def validate_without_origin(var: Any, ann: Any) -> None:
  ann_to_test = ann if ann != float else (int, float)

  if ann_to_test == Any: return

  if not isinstance(var, ann_to_test):
    raise TypeError(format_type_error(var, ann))
  
def validate_with_origin(origin: Any, var: Any, args: tuple) -> None:
    if origin == Union:
      return validate_union(var, args)
    if origin == Literal:
      return validate_literal(var, args)
    if origin == list:
      return validate_list(var, args[0])
    if origin == tuple:
      return validate_tuple(var, args)
    if origin == dict:
      return validate_dict(var, args)

    print('VALIDATION WITH ORIGIN NOT IMPLEMENTED:', var, args, origin)

def validate_dict(var: Any, anns: tuple) -> None:
  if not isinstance(var, dict):
    raise TypeError(format_type_error(var, dict))

  key_class = anns[0]
  val_class = anns[1]
  for key in var:
    try:
      validate_var(key, key_class)
    except Exception as e:
      raise TypeError(f'dict key ' + e.args[0])

    try:
      validate_var(var[key], val_class)
    except Exception as e:
      raise TypeError(f'dict key ({key}) -> value ' + e.args[0])

def validate_literal(var: Any, options: tuple) -> None:
  if var not in options:
    raise TypeError(f'({var}) is not one of the valid options: ({options})')

def validate_union(var: Any, anns: tuple) -> None:
  for ann in anns:
    try:
      validate_var(var, ann)
      return
    except:
      pass
  
  anns_str = format_ann_args(anns, 'union')
  error_message = f'({var}) is ({type(var).__name__}), but should be ({anns_str})'
  raise TypeError(error_message)

def validate_tuple(var: Any, anns: tuple) -> None:
  if not isinstance(var, tuple):
    raise TypeError(format_type_error(var, tuple))
  
  var_len = len(var)
  anns_len = len(anns)
  if var_len != anns_len:
    raise TypeError(f'tuple with length {var_len}, should have length {anns_len}')

  for i in range(anns_len):
    try:
      validate_var(var[i], anns[i])
    except Exception as e:
      raise(TypeError(f'error in position ({i}): ' + e.args[0]))

def validate_list(var: Any, ann: Any) -> None:
  if not isinstance(var, list):
    raise TypeError(format_type_error(var, list))
  
  for i in range(len(var)):
    try:
      validate_var(var[i], ann)
    except Exception as e:
      raise TypeError(f'error in position ({i}): '+e.args[0])

def format_type_error(var, ann) -> str:
  return f'({var}) is ({type(var).__name__}), but should be ({ann.__name__})'

def format_ann_args(ann_args, origin_type: str) -> str:
  if not isinstance(ann_args, tuple):
    return ann_args.__name__

  anns_str = ''
  count_anns = len(ann_args)
  for i in range(count_anns):
    ann_arg = ann_args[i]
    origin = get_origin(ann_arg)
    anns_str_prefix = ann_arg.__name__

    if origin is not None:
      if origin == tuple:
        anns_str_prefix += '[' + format_ann_args(get_args(ann_arg), 'tuple')+ ']'
      if origin == list:
        anns_str_prefix += '[' + get_args(ann_arg)[0].__name__ + ']'
      if origin == dict:
        anns_str_prefix += '[' + format_ann_args(get_args(ann_arg), 'dict')+ ']'
      # if origin == Union:
      #   anns_str_prefix = format_ann_args(get_args(ann_arg), 'union')


    if i == count_anns-1:
      anns_str += anns_str_prefix
    else:
      if origin_type == 'union':
        anns_str += anns_str_prefix + ' | '
      # elif origin_type == 'tuple':
      else:
        anns_str += anns_str_prefix + ', '

  return anns_str


from typing import get_origin, get_args, Union, Literal, Any
from copy import deepcopy


def from_dict(cls, d: dict):
  if not isinstance(d, dict): raise TypeError('Input is not a dict!')
  d_copy = deepcopy(d)

  dc_fields = cls.__dataclass_fields__  # type: ignore
  for key in d_copy:
    if key not in dc_fields:
      raise TypeError(f'{key} is not an attribute of {cls.__name__}')

    d_copy[key] = gen_from_any(d_copy[key], dc_fields[key].type)
    
  return cls(**d_copy) # type: ignore

def gen_from_any(d_val, ann):
  origin = get_origin(ann)
  if origin is not None:
    if origin == list:
      return gen_from_list(d_val, get_args(ann)[0])
    if origin == Union:
      return gen_from_union(d_val, get_args(ann))
    if origin == Literal: #(int, bool, str, bytes, None)
      return d_val
    if origin == tuple:
      return gen_from_tuple(d_val, get_args(ann))
    if origin == dict:
      return gen_from_dict(d_val, get_args(ann))

    print('GENERATION WITH ORIGIN NOT IMPLEMENTED:', ann, origin, d_val)
    return d_val
  else:
    if isinstance(d_val, ann):
      return d_val
    if ann == Any:
      return d_val

    # check if has from_dict class method
    from_dict_method = getattr(ann, 'from_dict', None)
    if from_dict_method is not None:
      return from_dict_method(d_val)
    
    try:
      return ann(**d_val)
    except:
      return d_val

def gen_from_list(d_val, ann):
  if not isinstance(d_val, list): return d_val

  gen_list = []
  for obj in d_val:
    gen_list.append(gen_from_any(obj, ann))
  
  return gen_list

def gen_from_tuple(d_val, anns):
  if not isinstance(d_val, (tuple, list)): return d_val

  anns_len = len(anns)
  d_val_len = len(d_val)

  if anns_len != d_val_len: return d_val

  gen_lis = []
  for i in range(len(anns)):
    gen_lis.append(gen_from_any(d_val[i], anns[i]))

  return tuple(gen_lis)

def gen_from_union(d_val, anns):
  for ann in anns:
    try:
      new_val = gen_from_any(d_val, ann)
    except:
      new_val = d_val
    
    if d_val != new_val:
      return new_val
  
  return d_val

def gen_from_dict(d_val, amns):
  if not isinstance(d_val, dict): return d_val
  new_dict = {}
  key_class = amns[0]
  val_class = amns[1]

  for val in d_val:
    new_key = gen_from_any(val, key_class) #needs to be hashable (int, bool, str, tuple)
    new_val = gen_from_any(d_val[val], val_class)
    try:
      new_dict[new_key] = new_val
    except:
      return d_val

  return new_dict